package builtins

import (
	"context"
	"ucl.lmika.dev/ucl"
)

func Itrs() ucl.Module {
	return ucl.Module{
		Name: "itrs",
		Builtins: map[string]ucl.BuiltinHandler{
			"from":    iterFrom,
			"tap":     iterTap,
			"to-list": iterToList,
		},
	}
}

func iterFrom(ctx context.Context, args ucl.CallArgs) (any, error) {
	var listable ucl.Listable

	if err := args.Bind(&listable); err != nil {
		return nil, err
	}

	return &fromIterator{listable: listable}, nil
}

type fromIterator struct {
	idx      int
	listable ucl.Listable
}

func (f *fromIterator) String() string {
	return "fromIterator{}"
}

func (f *fromIterator) HasNext() bool {
	return f.idx < f.listable.Len()
}

func (f *fromIterator) Next(ctx context.Context) (ucl.Object, error) {
	if f.idx >= f.listable.Len() {
		return nil, nil
	}
	v := f.listable.Index(f.idx)
	f.idx++
	return v, nil
}

func iterTap(ctx context.Context, args ucl.CallArgs) (any, error) {
	var (
		src ucl.Iterable
		tap ucl.Invokable
	)

	if err := args.Bind(&src, &tap); err != nil {
		return nil, err
	}

	return &tappedIterator{src, tap}, nil
}

type tappedIterator struct {
	src ucl.Iterable
	tap ucl.Invokable
}

func (f *tappedIterator) String() string {
	return "tappedIterator{}"
}

func (f *tappedIterator) HasNext() bool {
	return f.src.HasNext()
}

func (f *tappedIterator) Next(ctx context.Context) (ucl.Object, error) {
	v, err := f.src.Next(ctx)
	if err != nil {
		return nil, err
	}

	if _, err := f.tap.Invoke(ctx, v); err != nil {
		return nil, err
	}

	return v, nil
}

func iterToList(ctx context.Context, args ucl.CallArgs) (any, error) {
	var itr ucl.Iterable
	if err := args.Bind(&itr); err != nil {
		return nil, err
	}

	target := ucl.NewListObject()
	for itr.HasNext() {
		v, err := itr.Next(ctx)
		if err != nil {
			return nil, err
		}
		if err := target.Insert(-1, v); err != nil {
			return nil, err
		}
	}

	return target, nil
}
